/* David Anderson-Gould
 * CS113 Frost
 * October 22, 2007
 */
// Copyright (c) 2008, David Anderson-Gould
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of Deapuzzle nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY DAVID ANDERSON-GOULD ``AS IS'' AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL DAVID ANDERSON-GOULD BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import ucigame.*;

public class Deapuzzle extends Ucigame {

	Sprite ship;
	int shipRotation; // keep track of how many degrees we've rotated.

	Sprite ball;
	int bounces;
	boolean ballExists;
	Sprite ballLauncher; // keep track of position ball should start at

	// These sprites define the features of the level.
	int currentLevel;
	final int lastLevel = 1;
	Sprite shipWall;
	Sprite ballWall;
	Sprite goal;
	Sprite goalSuccess;

	boolean victory;
	Sprite victoryScreen;

	int[] shipStartingX = { 189, 189, 0 };
	int[] shipStartingY = { 335, 226, 0 };

	int[] goalX = { 425, 450, 0 };
	int[] goalY = { 215, 215, 0 };

	Sprite rearFire; // flame out the back of the ship.

	final int shipCenterX = 22;
	final int shipCenterY = 28;

	public void setup()
	{

		window.size(640, 480);
		window.title("Deapuzzle");
		framerate(30);
		window.showFPS();

		canvas.background(0, 0, 0);

		doLocalSetup();

	}

	// This is a separate method so that the game can be restarted.
	public void doLocalSetup()
	{

		// set up the ship images.
		ship = makeSprite(getImage("ship-60tall.png", 255, 0, 0));

		rearFire = makeSprite(97, 60);
		rearFire.addFrames(
				getImage("rearfire-mask.png", 255, 0, 0),
				0, 0,
				97, 0,
				0, 60,
				97, 60);
		rearFire.framerate(10);

		// this sprite lets me easily find the right spot to launch the ball
		// from, no matter the ship's orientation.
		ballLauncher = makeSprite(getImage("blank1.png", 255, 0, 0));
		ship.pin(ballLauncher, 65, 28);

		// pin fire to ship.
		ship.pin(rearFire, 0 - ship.width(), 0);
		rearFire.hide();

		// set up ball.
		ball = makeSprite(getImage("bullet-40.png", 255, 0, 0));
		ballExists = false;


		// set up goal images.
		goal = makeSprite(getImage("goal-red.png", 0, 0, 255));
		goalSuccess = makeSprite(getImage("goal-green.png", 0, 0, 255));
		goal.pin(goalSuccess, 0, 0);
		goalSuccess.hide();

		// set up victory image.
		victoryScreen = makeSprite(getImage("victorymessage.png"));
		victoryScreen.position(canvas.width()/2 - victoryScreen.x()/2,
				canvas.height()/2 - victoryScreen.y()/2);

		setUpCurrentLevel();

	}

	public void draw()
	{

		canvas.clear();

		if (victory)
		{
			victoryScreen.draw();
			return;
		}

		ship.move();
		ship.rotate(shipRotation, shipCenterX, shipCenterY);
		ship.draw();

		ballLauncher.draw();

		ship.stopIfCollidesWith(shipWall, PIXELPERFECT);

		keepUnderMaxSpeed(ship, 10.0);

		rearFire.hide();

		shipWall.draw();

		// ball goes here.
		if (ballExists)
		{
			ball.move();
			ball.draw();

			ball.bounceIfCollidesWith(ballWall, ship, PIXELPERFECT);

			if (ball.collided())
				bounces++;

			if (bounces > 5)
			{
				bounces = 0;
				ballExists = false;
			}

			ball.checkIfCollidesWith(goal, PIXELPERFECT);

			if (ball.collided())
			{
				goalSuccess.show();

				ballExists = false;
				bounces = 0;

				if (currentLevel < lastLevel)
					startTimer("nextLevel", 2500);
				else
					victory = true;
			}

			keepUnderMaxSpeed(ball, 15.0);
		}

		goal.draw();

		ballWall.draw();
	}

	public void onKeyPress()
	{
		// w/up fires the main thruster.
		if (keyboard.key() == keyboard.UP || keyboard.key() == keyboard.W)
		{
			rearFire.show(); // show the player pretty flames.

			// accelerate the ship in the direction it's facing...
			// y accel proportional to sin (theta)
			// x accel proportional to cos (theta)

			// convert to radians...
			double thetaRad = 3.14159 * (double)shipRotation / 180.0;

			ship.motion(0.1 * Math.cos(thetaRad),
						0.1 * Math.sin(thetaRad),
						ADD);
		}
		// s/down moves the ship to full stop.
		if (keyboard.key() == keyboard.DOWN || keyboard.key() == keyboard.S)
		{
			ship.motion(0.8, 0.8, MULTIPLY);
		}
		// a/left rotates ship left.
		if (keyboard.key() == keyboard.LEFT || keyboard.key() == keyboard.A)
			shipRotation -= 5;
		// d/right rotates ship right.
		if (keyboard.key() == keyboard.RIGHT || keyboard.key() == keyboard.D)
			shipRotation += 5;

		// keep rotation 0 <= x <= 359
		if (shipRotation < 0)
			shipRotation += 360;
		else if (shipRotation > 359)
			shipRotation -= 360;

		// space fires the ball, or, if the game is over, restarts the game
		if (keyboard.key() == keyboard.SPACE && !ballExists)
		{
			if (victory)
			{
				victory = false;

				doLocalSetup();
				return;
			}

			ballExists = true;

			double thetaRad = 3.14159 * (double)shipRotation / 180.0;

			int x = ship.xPixel() + shipCenterX;
			int y = ship.yPixel() + shipCenterY;

			x += (int)(42 * Math.cos(thetaRad));
			y += (int)(42 * Math.sin(thetaRad));

			x -= 20;
			y -= 20;

			ball.position(x, y);


			// ball gets initial velocity of ship plus a boost in the direction the ship
			// is pointed.
			ball.motion(ship.xspeed(), ship.yspeed());

			ball.motion(7.5 * Math.cos(thetaRad),
						7.5 * Math.sin(thetaRad),
						ADD);
		}

		// display debug info...
		if (keyboard.key() == keyboard.BACKQUOTE)
		{
			System.out.format("rotation %d; theta %.3f; cos %.3f; sin %.3f%n",
				shipRotation,
					3.14159 * (double)shipRotation/180.0,
					Math.cos(3.14159 * (double)shipRotation/180.0),
					Math.sin(3.14159 * (double)shipRotation/180.0));


			System.out.format("ship pos: %d, %d; ship veloc: %.3f, %.3f%n",
					ship.x(), ship.y(), ship.xspeed(), ship.yspeed());

			ship.rotate(shipRotation, shipCenterX, shipCenterY);

			System.out.format("post rotate ship pos: %d, %d",
					ship.x(), ship.y());
		}
	}

	private void setUpCurrentLevel()
	{
		shipWall = makeSprite(getImage("shipWall" + currentLevel + ".png", 255, 0, 0));
		ballWall = makeSprite(getImage("ballWall" + currentLevel + ".png", 0, 0, 255));

		ship.position(shipStartingX[currentLevel], shipStartingY[currentLevel]);
		shipRotation = 0;

		goal.position(goalX[currentLevel], goalY[currentLevel]);
		goalSuccess.hide();



	}

	private void keepUnderMaxSpeed(Sprite s, double maxSpeed)
	{
		double xSpeed = s.xspeed();
		double ySpeed = s.yspeed();

		if ((xSpeed * xSpeed) + (ySpeed * ySpeed) > maxSpeed * maxSpeed)
		{
			// Object is moving too fast.
			// Scale its speed down while keeping the same direction.

			// derive theta via tan theta = opposite (y side) / adjacent (x side).
			double theta = Math.atan(ySpeed / xSpeed);

			s.motion(maxSpeed * Math.cos(theta), maxSpeed * Math.sin(theta));

		}
	}

	public void nextLevelTimer()
	{
		currentLevel++;

		stopTimer("nextLevel");


		if (currentLevel <= lastLevel)
			setUpCurrentLevel();

	}
}
